/*
 *	Quarticurve KWin client
 *
 *	Copyright (C) 1999, 2001 Daniel Duley <mosfet@kde.org>
 *	Matthias Ettrich <ettrich@kde.org>
 *	Karol Szwed <gallium@kde.org>
 *	Than Ngo <than@redhat.com>
 *	Copyright (C) 2007-2008 Kevin Kofler <Kevin@tigcc.ticalc.org> (KDE 4 port)
 *	Copyright (C) 1999 Daniel M. Duley <mosfet@kde.org> (kColorBitmaps)
 *
 *	Draws mini titlebars for tool windows.
 *	Many features are now customizable.
 */

#include "quarticurveclient.h"
//Added by qt3to4:
#include <Q3HBoxLayout>
#include <QPixmap>
#include <QResizeEvent>
#include <QMouseEvent>
#include <QEvent>
#include <Q3VBoxLayout>
#include <QShowEvent>
#include <QPaintEvent>
#include <Q3Painter>

#include <kconfig.h>
#include <kconfiggroup.h>
#include <kglobal.h>
#include <qimageblitz/qimageblitz.h>
#include <klocale.h>
#include <qlayout.h>
#include <qdrawutil.h>
#include <qbitmap.h>
#include <qimage.h>
#include <qtooltip.h>
#include <qapplication.h>
#include <qlabel.h>
#include <kdebug.h>


#define BASE_BUTTON_SIZE  16
#define BORDER_WIDTH      6
#define CORNER_RADIUS     12

#define BUTTON_DIAM       12
#define TOP_GRABBAR_WIDTH 2
#define BOTTOM_CORNER     5


namespace Quarticurve
{
#include "bitmaps.h"

QPixmap* titlePix;
QPixmap* titleBuffer;
QPixmap* aUpperGradient;
QPixmap* iUpperGradient;

QPixmap* pinDownPix;
QPixmap* pinUpPix;
QPixmap* ipinDownPix;
QPixmap* ipinUpPix;

QPixmap* btnUpPix;
QPixmap* btnDownPix;
QPixmap* ibtnUpPix;
QPixmap* ibtnDownPix;

QPixmap* bottomLeftPix;
QPixmap* bottomRightPix;
QPixmap* abottomLeftPix;
QPixmap* abottomRightPix;

QuarticurveHandler* clientHandler;

bool Quarticurve_initialized = false;
bool useGradients;
bool showGrabBar;
bool showTitleBarStipple;
bool largeToolButtons;

static int grabBorderWidth;
static int borderWidth;
static int toolTitleHeight;
static int normalTitleHeight;

static void kColorBitmaps(QPainter *p, const QPalette &pal, int x, int y, int w,
                          int h, bool isXBitmaps, const uchar *lightColor,
                          const uchar *midColor, const uchar *midlightColor,
                          const uchar *darkColor, const uchar *blackColor,
                          const uchar *whiteColor)
{
    const uchar *data[]={lightColor, midColor, midlightColor, darkColor,
        blackColor, whiteColor};

    QColor colors[]={pal.color(QPalette::Light), pal.color(QPalette::Mid),
        pal.color(QPalette::Midlight), pal.color(QPalette::Dark),
        Qt::black, Qt::white};

    int i;
    QBitmap b;
    for(i=0; i < 6; ++i){
        if(data[i]){
            QSize s(w,h);
            b = QBitmap::fromData(s, data[i], isXBitmaps ? QImage::Format_MonoLSB : QImage::Format_Mono );
            b.setMask(b);
            p->setPen(colors[i]);
            p->drawPixmap(x, y, b);
        }
    }
}

static QPixmap& pixmapGradient(QPixmap &pixmap, const QColor &ca,
                               const QColor &cb, Blitz::GradientType eff)
{
    QImage image = Blitz::gradient(pixmap.size(), ca, cb, eff);
    pixmap = QPixmap::fromImage(image);
    return pixmap;
}

static QPixmap& pixmapIntensity(QPixmap &pixmap, float percent)
{
    QImage image = pixmap.toImage();
    Blitz::intensity(image, percent);
    pixmap = QPixmap::fromImage(image);

    return pixmap;
}

QuarticurveHandler::QuarticurveHandler()
{
	readConfig();
	createPixmaps();
	Quarticurve_initialized = true;
}


QuarticurveHandler::~QuarticurveHandler()
{
	Quarticurve_initialized = false;
	freePixmaps();
}


KDecoration* QuarticurveHandler::createDecoration( KDecorationBridge* bridge )
{
	return new QuarticurveClient( bridge, this );
}


bool QuarticurveHandler::reset(unsigned long changed)
{
	Quarticurve_initialized = false;
	freePixmaps();
	readConfig();
	createPixmaps();
	Quarticurve_initialized = true;
	bool needHardReset = true;
	if (changed & SettingColors)
		needHardReset = false;

	if (needHardReset) {
		return true;
	} else {
		resetDecorations(changed);
		return false;
	}

	return true;
}


void QuarticurveHandler::readConfig()
{
	KSharedConfigPtr conf = KGlobal::config();
	KConfigGroup group(conf, "Quarticurve");

	showGrabBar = group.readEntry("ShowGrabBar", true);
	showTitleBarStipple = group.readEntry("ShowTitleBarStipple", true);
	useGradients = group.readEntry("UseGradients", true);
	int size = group.readEntry("TitleBarSize", 0);

	if (size < 0) size = 0;
	if (size > 2) size = 2;

	normalTitleHeight = BASE_BUTTON_SIZE + (4*size);
	toolTitleHeight = normalTitleHeight - 4;
	largeToolButtons = (toolTitleHeight >= 16) ? true : false;

	int new_borderWidth;
	switch(options()->preferredBorderSize(this)) {
	case BorderLarge:
		new_borderWidth = 8;
		break;
	case BorderVeryLarge:
		new_borderWidth = 12;
		break;
	case BorderHuge:
		new_borderWidth = 18;
		break;
	case BorderVeryHuge:
		new_borderWidth = 27;
		break;
	case BorderOversized:
		new_borderWidth = 40;
		break;
	case BorderTiny:
	case BorderNormal:
	default:
		new_borderWidth = 4;
	}

	borderWidth = new_borderWidth;
	grabBorderWidth = (borderWidth > 15) ? borderWidth + 15 : 2*borderWidth;
}


// This paints the button pixmaps upon loading the style.
void QuarticurveHandler::createPixmaps()
{
	// Make the titlebar stipple optional
	if (showTitleBarStipple)
	{
		Q3Painter maskPainter;
		int x, y;
		titlePix = new QPixmap();
		titlePix->resize(132, normalTitleHeight+2);
		titlePix->fill();
		QBitmap mask(132, normalTitleHeight+2);

		mask.fill(Qt::color0);

		maskPainter.begin(&mask);
		maskPainter.setPen(Qt::color1);

		QColor lighterColor(options()->color(ColorTitleBar, true).light (150));
		int h, s, v;
		lighterColor.hsv (&h, &s, &v);
		s /= 2;
		s = (s > 255) ? 255 : (int) s;
		QColor satColor(h, s, v, QColor::Hsv);

		pixmapGradient(*titlePix,
				satColor,
				satColor.dark(150),
				Blitz::VerticalGradient);

		for(y = 0; y < (normalTitleHeight+2); y++) {
			for(x = (3 - y) % 4; x < 132; x += 4) {
				maskPainter.drawPoint(x, y);
			}
		}

		maskPainter.end();
		titlePix->setMask(mask);
	} else
		titlePix = NULL;

	// Create titlebar gradient images if required
	aUpperGradient = NULL;
	iUpperGradient = NULL;

	// Set the sticky pin pixmaps;
	QColorGroup g;
	Q3Painter p;

	// Active pins
	g = options()->palette( ColorButtonBg, true );

	pinUpPix = new QPixmap();
	pinUpPix->resize(BASE_BUTTON_SIZE, BASE_BUTTON_SIZE);
	pinUpPix->fill();
	p.begin( pinUpPix );
	kColorBitmaps( &p, g, 0, 0, BASE_BUTTON_SIZE, BASE_BUTTON_SIZE, true, pinup_white_bits,
		pinup_gray_bits, NULL, NULL, pinup_dgray_bits, NULL );
	p.end();
	pinUpPix->setMask( QBitmap(BASE_BUTTON_SIZE, BASE_BUTTON_SIZE, pinup_mask_bits, true) );

	pinDownPix = new QPixmap();
	pinDownPix->resize(BASE_BUTTON_SIZE, BASE_BUTTON_SIZE);
	pinDownPix->fill();
	p.begin( pinDownPix );
	kColorBitmaps( &p, g, 0, 0, BASE_BUTTON_SIZE, BASE_BUTTON_SIZE, true, pindown_white_bits,
		pindown_gray_bits, NULL, NULL, pindown_dgray_bits, NULL );
	p.end();
	pinDownPix->setMask( QBitmap(BASE_BUTTON_SIZE, BASE_BUTTON_SIZE, pindown_mask_bits, true) );

	// Inactive pins
	g = options()->palette( ColorButtonBg, false );

	ipinUpPix = new QPixmap();
	ipinUpPix->resize(BASE_BUTTON_SIZE, BASE_BUTTON_SIZE);
	ipinUpPix->fill();
	p.begin( ipinUpPix );
	kColorBitmaps( &p, g, 0, 0, BASE_BUTTON_SIZE, BASE_BUTTON_SIZE, true, pinup_white_bits,
		pinup_gray_bits, NULL, NULL, pinup_dgray_bits, NULL );
	p.end();
	ipinUpPix->setMask( QBitmap(BASE_BUTTON_SIZE, BASE_BUTTON_SIZE, pinup_mask_bits, true) );

	ipinDownPix = new QPixmap();
	ipinDownPix->resize(BASE_BUTTON_SIZE, BASE_BUTTON_SIZE);
	ipinDownPix->fill();
	p.begin( ipinDownPix );
	kColorBitmaps( &p, g, 0, 0, BASE_BUTTON_SIZE, BASE_BUTTON_SIZE, true, pindown_white_bits,
		pindown_gray_bits, NULL, NULL, pindown_dgray_bits, NULL );
	p.end();
	ipinDownPix->setMask( QBitmap(BASE_BUTTON_SIZE, BASE_BUTTON_SIZE, pindown_mask_bits, true) );

	// Create a title buffer for flicker-free painting
	titleBuffer = new QPixmap();

	// Cache all possible button states

	btnUpPix = new QPixmap();
	btnUpPix->resize(BASE_BUTTON_SIZE, BASE_BUTTON_SIZE);
	btnDownPix = new QPixmap();
	btnDownPix->resize(BASE_BUTTON_SIZE, BASE_BUTTON_SIZE);
	ibtnUpPix = new QPixmap();
	ibtnUpPix->resize(BASE_BUTTON_SIZE, BASE_BUTTON_SIZE);
	ibtnDownPix = new QPixmap();
	ibtnDownPix->resize(BASE_BUTTON_SIZE, BASE_BUTTON_SIZE);

	// Draw the button state pixmaps
	g = options()->palette( ColorButtonBg, true );

	drawButtonBackground( btnUpPix, g, false, true );
	drawButtonBackground( btnDownPix, g, true, true );

	g = options()->palette( ColorButtonBg, false );

	drawButtonBackground( ibtnUpPix, g, false, false );
	drawButtonBackground( ibtnDownPix, g, true, false );

	QImage bottomleft(bottom_left_xpm);
	QImage bottomright(bottom_right_xpm);
	QImage abottomleft(bottom_left_xpm);
	QImage abottomright(bottom_right_xpm);

	recolor(bottomleft, options()->color( ColorTitleBar, false ).light(95));
	recolor(bottomright, options()->color( ColorTitleBar, false ).light(95));
	recolor(abottomleft, options()->color( ColorTitleBar, true ).light(135));
	recolor(abottomright, options()->color( ColorTitleBar, true ).light(135));

	bottomLeftPix 	= new QPixmap();
	bottomRightPix	= new QPixmap();
	abottomLeftPix	= new QPixmap();	
	abottomRightPix	= new QPixmap();
	bottomLeftPix->convertFromImage(bottomleft);
	bottomRightPix->convertFromImage(bottomright);
	abottomLeftPix->convertFromImage(abottomleft);
	abottomRightPix->convertFromImage(abottomright);
}

// This is the recoloring method from the Keramik widget style,
// copyright (c) 2002 Malte Starostik <malte@kde.org>.
// Modified to work with 8bpp images.
void QuarticurveHandler::recolor( QImage &img, const QColor& color )
{
	int hue = -1, sat = 0, val = 228;
	if ( color.isValid() )
		color.hsv( &hue, &sat, &val );
	register int pixels = (img.depth() > 8 ? img.width() * img.height() : img.numColors());
	QVector<QRgb> vec;
	register Q_UINT32* data = ( img.depth() > 8 ? reinterpret_cast< Q_UINT32* >( img.bits() ) :
		reinterpret_cast< Q_UINT32* >( (vec = img.colorTable()).data() ) );
	
	for ( int i = 0; i < pixels; i++ )
	{
		QColor c( *data );
		int h, s, v;
		c.hsv( &h, &s, &v );
		h = hue;
		s = sat;
		v = v * val / 145;
		c.setHsv( h, QMIN( s, 255 ), QMIN( v, 255 ) );
		*data = ( c.rgb() & RGB_MASK ) | ( *data & ~RGB_MASK );
		data++;
	}

	if (img.depth() <= 8)
		img.setColorTable(vec);
}


void QuarticurveHandler::freePixmaps()
{
	// Free button pixmaps
	if (btnUpPix)
		delete btnUpPix;
	if(btnDownPix)
		delete btnDownPix;
	if (ibtnUpPix)
		delete ibtnUpPix;
	if (ibtnDownPix)
		delete ibtnDownPix;

	// Title images
	if (titleBuffer)
		delete titleBuffer;
	if (titlePix)
		delete titlePix;
	if (aUpperGradient)
		delete aUpperGradient;
	if (iUpperGradient)
		delete iUpperGradient;

	// Sticky pin images
	if (pinUpPix)
		delete pinUpPix;
	if (ipinUpPix)
		delete ipinUpPix;
	if (pinDownPix)
		delete pinDownPix;
	if (ipinDownPix)
		delete ipinDownPix;
}


void QuarticurveHandler::drawButtonBackground(QPixmap *pix, 
		const QColorGroup &g, bool sunken, bool active)
{
	Q3Painter p;

	bool highcolor = useGradients && (QPixmap::defaultDepth() > 8);
	QColor c = g.background();

	// Fill the background with a gradient if possible
	if (highcolor)
	{
		if (active)
		{
			pixmapGradient(*pix, c, Qt::white,
				Blitz::DiagonalGradient);
		} else
		{
			QColor inactiveTitleColor1(options()->color(ColorTitleBar, false));
			QColor inactiveTitleColor2(options()->color(ColorTitleBlend, false));
			pixmapGradient(*pix,
				inactiveTitleColor2,
				inactiveTitleColor1,
				Blitz::VerticalGradient);
		}
	} else
		pix->fill(c);

	p.begin(pix);
	p.setPen(sunken ? g.dark() : g.mid());
}


bool QuarticurveHandler::supports(KDecorationFactory::Ability) const
{
	return false;
}


QuarticurveButton::QuarticurveButton(QuarticurveClient *parent, const char *name,
		bool largeButton, int bpos, bool isOnAllDesktopsButton,
		const unsigned char *bitmap, const QString& tip, const int realizeBtns)
		: Q3Button(parent->widget(), name)
{
	realizeButtons = realizeBtns;
	setBackgroundMode( Qt::NoBackground );
	setToggleButton( isOnAllDesktopsButton );

	isMouseOver = false;
	deco = NULL;
	large = largeButton;
	isOnAllDesktops = isOnAllDesktopsButton;
	client = parent;

	pos = bpos;

	setFixedSize(BASE_BUTTON_SIZE, BASE_BUTTON_SIZE);

	if (bitmap)
		setBitmap(bitmap);

	QToolTip::add(this, tip);
}


QuarticurveButton::~QuarticurveButton()
{
	if (deco)
		delete deco;
}


QSize QuarticurveButton::sizeHint() const
{
	return( QSize(BASE_BUTTON_SIZE,BASE_BUTTON_SIZE ));
}


void QuarticurveButton::resizeEvent( QResizeEvent* e)
{
	doShape();
	Q3Button::resizeEvent(e);
}

void QuarticurveButton::showEvent(QShowEvent *)
{
	doShape();
}

void QuarticurveButton::doShape()
{
	// Obtain widget bounds.
	int w  = rect() .width();
	int h  = rect().height();
	int r = BUTTON_DIAM / 2;
	int dm = BUTTON_DIAM;
	QBitmap mask(size(), true);

	QPainter p3(&mask);
	QBrush blackbr(Qt::color1);
	p3.fillRect(0,0,w,h,blackbr);

	p3.setPen(Qt::color1);
	p3.setBrush(Qt::color1);
	if (pos == ButtonLeft) {
		p3.eraseRect(0, -TOP_GRABBAR_WIDTH, r, r);
		p3.drawPie(0, -TOP_GRABBAR_WIDTH, dm-1, dm-1, 90*16, 90*16);
		p3.drawArc(0, -TOP_GRABBAR_WIDTH, dm-1, dm-1, 90*16, 90*16);
	} else if (pos == ButtonRight)
		{
			p3.eraseRect(w-r , -TOP_GRABBAR_WIDTH, r,r);
			p3.drawPie(w-dm, -TOP_GRABBAR_WIDTH, dm-1, dm-1, 0*16, 90*16);
			p3.drawArc(w-dm, -TOP_GRABBAR_WIDTH, dm-1, dm-1, 0*16, 90*16);
		}
	p3.end();
	setMask(mask);
}

void QuarticurveButton::setBitmap(const unsigned char *bitmap)
{
	if (deco)
		delete deco;

	deco = new QBitmap(14, 14, bitmap, true);
	deco->setMask( *deco );
	repaint( false );
}


void QuarticurveButton::drawButton(QPainter *p)
{
	if (!Quarticurve_initialized)
		return;

	if (deco || isOnAllDesktops)
	{
		// Fill the button background with an appropriate button image
		QPixmap btnbg;

		if (isDown())
			btnbg = client->isActive() ? *btnDownPix : *ibtnDownPix;
		else
			btnbg = client->isActive() ? *btnUpPix : *ibtnUpPix;
      
		if (isMouseOver)
			pixmapIntensity(btnbg, 0.8);

		// Scale the background image if required
		// This is slow, but we assume this isn't done too often
		if ( !large )
		{
			btnbg.detach();
			btnbg.convertFromImage(btnbg.convertToImage().smoothScale(14, 14));
		}

		p->drawPixmap( 0, 0, btnbg );

	}

	// If we have a decoration bitmap, then draw that
	// otherwise we paint a menu button (with mini icon), or a sticky button.
	if ( deco )
	{
		// Select the appropriate button decoration color
		bool darkDeco = qGray( KDecoration::options()->color(
				KDecoration::ColorButtonBg,
				client->isActive()).rgb() ) > 127;

		QColor bgc = KDecoration::options()->color(KDecoration::ColorTitleBar, client->isActive());
		if (isMouseOver)
			p->setPen( darkDeco ? bgc.dark(120) : bgc.light(120) );
		else
			p->setPen( darkDeco ? bgc.dark(150) : bgc.light(150) );

		int xOff = (width()-14)/2;
		int yOff = (height()-14)/2;
		p->drawPixmap(isDown() ? xOff+1: xOff, isDown() ? yOff+1 : yOff, *deco);
	} else
	{
		QPixmap btnpix;

		if (isOnAllDesktops)
		{
			if (client->isActive())
				btnpix = isOn() ? *pinDownPix : *pinUpPix;
			else
				btnpix = isOn() ? *ipinDownPix : *ipinUpPix;
		} else
		{
			btnpix = client->icon().pixmap( QIcon::Small, QIcon::Normal );
        }
      
		// Intensify the image if required
		if (isMouseOver)
			btnpix = pixmapIntensity(btnpix, 0.8);
      
		// Smooth scale the pixmap for small titlebars
		// This is slow, but we assume this isn't done too often
		if ( !large )
			btnpix.convertFromImage(btnpix.convertToImage().smoothScale(14, 14));
      
		p->drawPixmap( 0, 0, btnpix );
	}

	QColorGroup g;
	p->setPen(g.dark());
}


// Make the protected member public
void QuarticurveButton::turnOn( bool isOn )
{
	if ( isToggleButton() )
		setOn( isOn );
}


void QuarticurveButton::enterEvent(QEvent *e) 
{ 
	isMouseOver=true;
	repaint(false); 
	Q3Button::enterEvent(e);
}


void QuarticurveButton::leaveEvent(QEvent *e)
{ 
	isMouseOver=false;
	repaint(false); 
	Q3Button::leaveEvent(e);
}


void QuarticurveButton::mousePressEvent( QMouseEvent* e )
{
	last_button = e->button();
	QMouseEvent me( e->type(), e->pos(), e->globalPos(),
			(e->button()&realizeButtons)?Qt::LeftButton:Qt::NoButton, e->state() );
	Q3Button::mousePressEvent( &me );
}


void QuarticurveButton::mouseReleaseEvent( QMouseEvent* e )
{
	last_button = e->button();
	QMouseEvent me( e->type(), e->pos(), e->globalPos(),
			(e->button()&realizeButtons)?Qt::LeftButton:Qt::NoButton, e->state() );
	Q3Button::mouseReleaseEvent( &me );
}


void QuarticurveButton::reset()
{
	// repaint the whole thing
	repaint(false);
}


void QuarticurveButton::setTipText(const QString &tip) {
	if ( KDecoration::options()->showTooltips() ) {
		QToolTip::remove(this );
		QToolTip::add(this, tip );
	}
}


QuarticurveClient::QuarticurveClient( KDecorationBridge* bridge, KDecorationFactory* factory )
		: KDecoration (bridge, factory)
{
}


void QuarticurveClient::init()
{
	createMainWidget( Qt::WResizeNoErase | Qt::WStaticContents | Qt::WRepaintNoErase );

	widget()->installEventFilter( this );
	widget()->setAttribute(Qt::WA_PaintOutsidePaintEvent,true);

	// No flicker thanks
	widget()->setBackgroundMode( Qt::NoBackground );

	// Set button pointers to NULL so we can track things
	for(int i=0; i < QuarticurveClient::BtnCount; i++)
		button[i] = NULL;

	// Finally, toolWindows look small
	if ( isTool() ) {
		titleHeight  = toolTitleHeight;
		largeButtons = false;
	} else {
		titleHeight  = normalTitleHeight;
		largeButtons = true;
	}

	// Pack the windowWrapper() window within a grid
	Q3VBoxLayout* g = new Q3VBoxLayout(widget());
	g->setResizeMode(QLayout::FreeResize);
	g->addSpacing(TOP_GRABBAR_WIDTH);       // Top grab bar

	// Pack the titlebar HBox with items
	hb = new Q3HBoxLayout();
	hb->setSpacing(0);
	hb->setMargin(0);
	hb->setResizeMode( QLayout::FreeResize );

	hb->addSpacing(2);
	addClientButtons( options()->titleButtonsLeft(), true );
	titlebar = new QSpacerItem( 10, titleHeight,
		QSizePolicy::Expanding, QSizePolicy::Minimum );
	hb->addItem(titlebar);

	//hb->addSpacing(2);
	addClientButtons( options()->titleButtonsRight(), false );
	hb->addSpacing(2);

	g->addLayout( hb );
	g->addSpacing(1); // line under titlebar

	// Add the middle section
	hb = new Q3HBoxLayout();
	hb->addSpacing(BORDER_WIDTH);
	if (isPreview())
		hb->addWidget(new QLabel( i18n( "<center><b>Quarticurve preview</b></center>" ), widget()));
	else
		hb->addWidget(new QLabel("", widget()));
	hb->addSpacing(BORDER_WIDTH);
	g->addLayout( hb );

	// Determine the size of the lower grab bar
	if ( showGrabBar && (!isTool()) )
		g->addSpacing(BORDER_WIDTH); // bottom handles
	else
		g->addSpacing(4); // bottom handles
}


bool QuarticurveClient::isTool() const
{
	NET::WindowType type = windowType(NET::NormalMask|NET::ToolbarMask|NET::UtilityMask|NET::MenuMask);
	return ((type==NET::Toolbar)||(type==NET::NET::Utility)||(type==NET::Menu));
}


void QuarticurveClient::addClientButtons( const QString& s, bool isLeft )
{
	int pos;
	// Make sure we place the spacing between the buttons
	bool first_button = true;
	QuarticurveButton* last_button = NULL;

	if (s.length() > 0) {
		for(int i = 0; i < s.length(); i++) {
			if (i == 0 && isLeft)
				pos = ButtonLeft;
			else
				pos = ButtonMid;

			switch( s[i].latin1() )
			{
				// Menu button
				case 'M':
					if (!button[BtnMenu])
					{
						button[BtnMenu] = new QuarticurveButton(this, "menu",
							largeButtons, pos, false, menu_bits, i18n("Menu"), Qt::LeftButton|Qt::RightButton);
						connect( button[BtnMenu], SIGNAL(pressed()),
							this, SLOT(menuButtonPressed()) );
						connect( button[BtnMenu], SIGNAL(released()),
							this, SLOT(menuButtonReleased()));
						if (! first_button)
							hb->addSpacing(2);
						else
							first_button = false;
						hb->addWidget( button[BtnMenu] );
						last_button = button[BtnMenu];
					}
					break;

				// Sticky button
				case 'S':
					if (!button[BtnOnAllDesktops])
					{
						button[BtnOnAllDesktops] = new QuarticurveButton(this, "on_all_desktops", 
							largeButtons, pos, true, NULL, i18n("On All Desktops"));
						button[BtnOnAllDesktops]->turnOn( isOnAllDesktops() );
						connect( button[BtnOnAllDesktops], SIGNAL(clicked()), 
							this, SLOT(toggleOnAllDesktops()) );
						hb->addSpacing(2);
						hb->addWidget( button[BtnOnAllDesktops] );
						last_button = button[BtnOnAllDesktops];
					}
					break;

				// Help button
				case 'H':
					if( providesContextHelp() && (!button[BtnHelp]) )
					{
						button[BtnHelp] = new QuarticurveButton(this, "help",
							largeButtons, pos, true, question_bits,
							i18n("Help"));
						connect( button[BtnHelp], SIGNAL( clicked() ),
							this, SLOT( showContextHelp() ));
						if (! first_button)
							hb->addSpacing(2);
						else
							first_button = false;
						hb->addWidget( button[BtnHelp] );
						last_button = button[BtnHelp];
					}
					break;

				// Minimize button
				case 'I':
					if ( (!button[BtnIconify]) && isMinimizable())
					{
						button[BtnIconify] = new QuarticurveButton(this, "iconify",
							largeButtons, pos, false, iconify_bits,
							i18n("Minimize"));
						connect( button[BtnIconify], SIGNAL( clicked()),
							this, SLOT(minimize()) );
						if (! first_button)
							hb->addSpacing(2);
						else
							first_button = false;
						hb->addWidget( button[BtnIconify] );
						last_button = button[BtnIconify];
					}
					break;

				// Maximize button
				case 'A':
				if ( (!button[BtnMax]) && isMaximizable())
				{
					button[BtnMax]  = new QuarticurveButton(this, "maximize",
						largeButtons, pos, false, maximize_bits,
						i18n("Maximize"), Qt::LeftButton|Qt::MidButton|Qt::RightButton);
					connect( button[BtnMax], SIGNAL( clicked()),
						this, SLOT(slotMaximize()) );
					if (! first_button)
						hb->addSpacing(2);
					else
						first_button = false;
					hb->addWidget( button[BtnMax] );
					last_button = button[BtnMax];
				}
				break;

				// Close button
				case 'X':
				if (!button[BtnClose])
				{
					button[BtnClose] = new QuarticurveButton(this, "close",
						largeButtons, pos, false, close_bits,
						i18n("Close"));
					connect( button[BtnClose], SIGNAL( clicked()),
						this, SLOT(closeWindow()) );
					if (! first_button)
						hb->addSpacing(2);
					else
						first_button = false;
					hb->addWidget( button[BtnClose] );
					last_button = button[BtnClose];
				}
				break;

				// Spacer item (only for non-tool windows)
				case '_':
				if ( !isTool() )
					hb->addSpacing(2);
			}
		}

		if (last_button) {
			if(not isLeft)
				last_button->pos = ButtonRight;
			else
				last_button->pos = LeftButtonRight;
		}
	}
}


void QuarticurveClient::iconChange()
{
	if (button[BtnMenu] && button[BtnMenu]->isVisible())
		button[BtnMenu]->repaint(false);
}


void QuarticurveClient::desktopChange()
{
	if (button[BtnOnAllDesktops]) {
		button[BtnOnAllDesktops]->turnOn(isOnAllDesktops());
		button[BtnOnAllDesktops]->repaint(false);
		button[BtnOnAllDesktops]->setTipText(isOnAllDesktops() ? i18n("Not On All Desktops") : i18n("On All Desktops"));
	}
}


void QuarticurveClient::slotMaximize()
{
	if (button[BtnMax])
	{
		switch (button[BtnMax]->last_button)
		{
			case Qt::MidButton:
				maximize(maximizeMode() ^ MaximizeVertical );
				break;
			case Qt::RightButton:
				maximize(maximizeMode() ^ MaximizeHorizontal );
				break;
			default:
				maximize(maximizeMode() == MaximizeFull ? MaximizeRestore : MaximizeFull );
		}
	}
}


void QuarticurveClient::resizeEvent( QResizeEvent* e)
{
	doShape();
	calcHiddenButtons();
  
	if (widget()->isVisibleToTLW())
	{
		widget()->update(widget()->rect());
		int dx = 0;
		int dy = 0;

		if ( e->oldSize().width() != width() )
			dx = 32 + QABS( e->oldSize().width() -  width() );

		if ( e->oldSize().height() != height() )
			dy = 8 + QABS( e->oldSize().height() -  height() );

		if ( dy )
			widget()->update( 0, height() - dy + 1, width(), dy );
		if ( dx )
		{
			widget()->update( width() - dx + 1, 0, dx, height() );
			widget()->update( QRect( QPoint(4,4), titlebar->geometry().bottomLeft() - QPoint(1,0) ) );
			widget()->update( QRect( titlebar->geometry().topRight(), QPoint(width() - 4,
				titlebar->geometry().bottom()) ) );
			// Titlebar needs no paint event
			// widget()->repaint(titlebar->geometry(), false);
			QApplication::postEvent( widget(), new QPaintEvent(titlebar->geometry()) );
		}
	}
}


void QuarticurveClient::captionChange()
{
	widget()->repaint( titlebar->geometry(), false );
}


void QuarticurveClient::paintEvent( QPaintEvent* )
{
	if (!Quarticurve_initialized)
		return;

	QColorGroup g;

	bool drawLeftDivider = true; 
	bool drawRightDivider = true; 

	// Obtain widget bounds.
	QRect r(widget()->rect());
	Q3Painter p(widget());

	int x = r.x();
	int y = r.y();
	int x2 = x + r.width() - 1;
	int y2 = y + r.height() - 1;
	int w  = r.width();
	int h  = r.height();
	g = options()->palette(ColorFrame, isActive());
	// Determine where to place the extended left titlebar
	//int leftFrameStart = (h > 42) ? y+titleHeight+26: y+titleHeight;

	// Determine where to make the titlebar color transition
	r = titlebar->geometry();
	//int rightOffset = r.x()+r.width()+1;

	QColor c2 = options()->color(ColorFrame, isActive() );
	// Fill with frame color behind RHS buttons
	///  p.fillRect( x, y+2, x2, titleHeight+1, c2);

	// Create a disposable pixmap buffer for the titlebar
	// very early before drawing begins so there is no lag
	// during painting pixels.
	titleBuffer->resize( w, titleHeight + TOP_GRABBAR_WIDTH );


	// Draw the title bar.
	r = titlebar->geometry();

	// Obtain titlebar blend colours
	QColor c1 = options()->color(ColorTitleBar, isActive() );

	Q3Painter p2( titleBuffer );
	p2.initFrom(widget());
	QColor activeTitleColor1(options()->color(ColorTitleBar,      true));
	QColor activeTitleColor2(options()->color(ColorTitleBlend,    true));

	QColor inactiveTitleColor1(options()->color(ColorTitleBar,    false));
	QColor inactiveTitleColor2(options()->color(ColorTitleBlend,  false));
	bool highcolor = useGradients && (QPixmap::defaultDepth() > 8);

	if (highcolor)
	{
		static QSize oldsize(0,0);
		QSize titleBufferSize(w, titleHeight + TOP_GRABBAR_WIDTH);
		if (oldsize != titleBufferSize)
		{
			oldsize = titleBufferSize;
			if (aUpperGradient)
			{
				delete aUpperGradient;
				aUpperGradient = NULL;
			}
			if (iUpperGradient)
			{
				delete iUpperGradient;
				iUpperGradient = NULL;
			}
	      
			// Create the titlebar gradients
			if (activeTitleColor1 != activeTitleColor2)
			{
				aUpperGradient = new QPixmap(oldsize);
				pixmapGradient(*aUpperGradient,
					activeTitleColor2,
					activeTitleColor1,
					Blitz::VerticalGradient);
			}

			if (inactiveTitleColor1 != inactiveTitleColor2)
			{
				iUpperGradient = new QPixmap(oldsize);
				pixmapGradient(*iUpperGradient,
					inactiveTitleColor2,
					inactiveTitleColor1,
					Blitz::VerticalGradient);
			}
		}
	}
  
	QPixmap* upperGradient = isActive() ? aUpperGradient : iUpperGradient;
	// Draw the titlebar gradient

	if (upperGradient)
		p2.drawPixmap(0, TOP_GRABBAR_WIDTH, *upperGradient);
	else
		p2.fillRect(0, TOP_GRABBAR_WIDTH, w, titleHeight, c1);

	QFont fnt = options()->font(true, true);

	if ( isTool() )
		fnt.setPointSize( fnt.pointSize()-2 );  // Shrink font by 2pt

	p2.setFont( fnt );

	// Draw the titlebar stipple if active and available
	if (isActive() && titlePix)
	{
		QFontMetrics fm(fnt);
		int captionWidth = fm.width(caption()) + 1;
		p2.drawTiledPixmap( r.x() + 2 + 2 + captionWidth, TOP_GRABBAR_WIDTH,
			r.width() - 2 - 4 - captionWidth, 
			titleHeight+1, *titlePix );
	}

	if (isActive())
	{
		p2.setPen( options()->color(ColorTitleBlend, isActive()).dark());
		p2.drawText(r.x() + 2 + 1, TOP_GRABBAR_WIDTH + 1,
			r.width() - 2 - 1, r.height(),
			Qt::AlignLeft | Qt::AlignVCenter, caption() );
	}

	p2.setPen( options()->color(ColorFont, isActive()) );
	p2.drawText(r.x() + 2, TOP_GRABBAR_WIDTH,
		r.width() - 2, r.height(),
		Qt::AlignLeft | Qt::AlignVCenter, caption() );

	// Main Title Bar background area
	p2.setPen(Qt::white);
	p2.drawLine(x + 1, y + 1, x2 - 1, y + 1);
	// This is kind of broken...
	// We fill in the inner part of the circle here.  This is dependent on BUTTON_DIAM
	p2.drawLine(x + 1, y + 1, x + 1, y + TOP_GRABBAR_WIDTH + titleHeight);
	p2.drawLine(x + 2, y + 2, x + 3, y + 2);
	p2.drawLine(x + 2, y + 2, x + 2, y + 3);
	p2.drawLine(x + w - 2 , y + 1, x + w - 2, y + TOP_GRABBAR_WIDTH + titleHeight);
	p2.drawLine(x + w - 3, y + 2, x + w - 3, y + 5);
	p2.drawLine(x + w - 4, y + 2, x + w - 3, y + 2);

	if (isActive())
	{
		QColor lighterColor (options()->color(ColorTitleBar, true).light (150));
		p2.setPen (lighterColor);
		p2.drawLine (r.x(), 2, r.x() + r.width(), 2);
		int h, s, v;
		lighterColor.hsv (&h, &s, &v);
		s /= 2;
		s = (s > 255) ? 255 : (int) s;

		QColor satColor(h, s, v, QColor::Hsv);
		p2.setPen (satColor);
		p2.drawLine (r.x(), 1, r.x() + r.width() - 2, 1);
	}

	p2.setPen(Qt::white);
	if (isActive())
	{
		for (int i = 0; i < BtnCount; i ++)
		{
			if (button[i] == NULL)
				continue;
			if (!button[i]->isVisible())
			{
				if (button[i]->pos == ButtonRight)
					drawRightDivider = false;
				// FIXME: Should be LeftButtonLeft if we had it
				if (button[i]->pos == LeftButtonRight)
					drawLeftDivider = false;
				continue;
			}
			QRect buttonSize = button[i]->geometry ();
			p2.setPen(Qt::white);
			p2.drawLine (buttonSize.x() - 1, TOP_GRABBAR_WIDTH,
			buttonSize.x() - 1, TOP_GRABBAR_WIDTH + titleHeight);
			if (button[i]->pos == ButtonRight)
				continue;
			else if (button[i]->pos == LeftButtonRight)
				p2.setPen(g.mid().light(120));
			else
				p2.setPen(g.dark());
			p2.drawLine (buttonSize.x() + buttonSize.width(), TOP_GRABBAR_WIDTH - 1,
				buttonSize.x() + buttonSize.width(), TOP_GRABBAR_WIDTH + titleHeight);
		}
	}


	// Top Left Button Area
	if (drawLeftDivider)
	{
		if (isActive ())
			p2.setPen(options()->color(ColorTitleBar, true).dark (150));
		else
			p2.setPen(g.mid());
		p2.drawLine (r.x() , y + 1, r.x() , y + titleHeight + TOP_GRABBAR_WIDTH);
	}



	// Top Right Button Area
	if (drawRightDivider)
	{
		if (isActive ())
			p2.setPen(options()->color(ColorTitleBar, true).dark (150));
		else
			p2.setPen(g.mid());
		p2.drawLine (r.x() + r.width() - 2, y + 1,
			r.x() + r.width() - 2 , y + titleHeight + TOP_GRABBAR_WIDTH);
	}

	// Black outer line
	p2.setPen(Qt::black);
	p2.drawRect(0,0,w,h);
	p2.drawArc(x, y, BUTTON_DIAM, BUTTON_DIAM, 90*16, 90*16);
	p2.drawArc(x + w - BUTTON_DIAM , y, BUTTON_DIAM, BUTTON_DIAM, 0*16, 90*16);
	p2.end();

	int sideStart = titleHeight + TOP_GRABBAR_WIDTH + 1;

	// Draw the left and right sides
  
	// Fill the left side first
	qDrawShadePanel(&p,
		// We compensate for the top and bottom parts of the bevel
		// by drawing 1 pixel below and above the frame part
		x + 1, y + (sideStart - 1),
		BORDER_WIDTH - 1, h - (sideStart + 2),
		g, false, 1, &g.brush(QColorGroup::Background));

	// Right Side
	qDrawShadePanel(&p,
		x2 - (BORDER_WIDTH - 2), y + (sideStart - 1),
		BORDER_WIDTH - 2, h - (sideStart + 2),
		g, false, 1, &g.brush(QColorGroup::Background));

	p.setPen(g.dark());
	p.drawLine(x2 - (BORDER_WIDTH - 1), sideStart, x2 - (BORDER_WIDTH - 1), h - sideStart);

	// Draw the bottom
	qDrawShadePanel(&p,
		x, y2 - (BORDER_WIDTH - 2),
		w, (BORDER_WIDTH - 2),
		g, false, 1, &g.brush(QColorGroup::Background));
	p.setPen(g.dark());
	p.drawLine(x, y2 - (BORDER_WIDTH - 1), x2, y2 - (BORDER_WIDTH - 1));

	// Line above the app and below the title bar
	p.setPen(g.dark());
	p.drawLine(x, y + titleHeight + TOP_GRABBAR_WIDTH,
		x2, y + titleHeight + TOP_GRABBAR_WIDTH);
	p.end();

	bitBlt( widget(), 0, 0, titleBuffer );

	// Draw an outer black frame
	p.begin(widget());
	p.setPen(Qt::black);
	p.drawRect(0,0,w,h);

	// Put on the bottom corners
	p.drawPixmap(0, h - bottomLeftPix->height(),
		isActive() ? *abottomLeftPix : *bottomLeftPix);
	p.drawPixmap(w - bottomRightPix->width(), h - bottomRightPix->height(), 
		isActive() ? *abottomRightPix : *bottomRightPix);
	p.end();
}


void QuarticurveClient::shadeChange()
{ ; }


void QuarticurveClient::doShape()
{
	// Obtain widget bounds.
	QRect r(widget()->rect());
	int x = 0;
	int y = 0;
	//int x2 = width() - 1;
	//int y2 = height() - 1;
	int w  = width();
	int h  = height();

	int rad = BUTTON_DIAM / 2;
	int dm = BUTTON_DIAM;

	QBitmap mask(w+1, h+1, true);

	Q3Painter p(&mask);

	p.fillRect(x, y, w+1, h+1, Qt::color1);

	p.eraseRect(x, y, rad, rad);
	p.eraseRect(w-rad+1, 0, rad, rad);

	p.eraseRect(x, h-BOTTOM_CORNER, BOTTOM_CORNER, BOTTOM_CORNER);
	p.eraseRect(w-BOTTOM_CORNER, h-BOTTOM_CORNER, BOTTOM_CORNER, BOTTOM_CORNER);

	p.setPen(Qt::color1);
	p.setBrush(Qt::color1);

	p.drawPie(x, y, dm, dm, 90*16, 90*16);
	p.drawArc(x, y, dm, dm, 90*16, 90*16);

	p.drawPie(w-dm, 0, dm, dm, 0*16, 90*16);
	p.drawArc(w-dm, 0, dm, dm, 0*16, 90*16);

	p.drawPixmap(x, h - bottomLeftPix->height(), bottomLeftPix->mask());

	p.drawPixmap(w-bottomRightPix->width(), h - bottomRightPix->height(), 
		bottomRightPix->mask());
	p.fillRect(x+BOTTOM_CORNER, h - bottomLeftPix->height(),
		bottomLeftPix->width()-BOTTOM_CORNER,
		bottomLeftPix->height()-BOTTOM_CORNER,
		Qt::color1);

	p.fillRect(w-bottomRightPix->width(), h - bottomRightPix->height(), 
		bottomRightPix->width()-BOTTOM_CORNER,
		bottomRightPix->height()-BOTTOM_CORNER,
		Qt::color1);

	p.end();
	setMask(mask);
}


void QuarticurveClient::showEvent(QShowEvent *)
{
	calcHiddenButtons();
	doShape();
	widget()->show();
}


void QuarticurveClient::mouseDoubleClickEvent( QMouseEvent * e )
{
	if ( titlebar->geometry().contains( e->pos() ) )
		titlebarDblClickOperation();
}


void QuarticurveClient::maximizeChange()
{
	if (button[BtnMax]) {
		button[BtnMax]->setBitmap((maximizeMode()==MaximizeFull) ? minmax_bits : maximize_bits);
		button[BtnMax]->setTipText((maximizeMode()==MaximizeFull) ? i18n("Restore") : i18n("Maximize"));
	}
}


void QuarticurveClient::borders( int& left, int& right, int& top, int& bottom ) const
{
	left = right = borderWidth;
	top = titleHeight + 4;
	bottom = (showGrabBar && isResizable()) ? grabBorderWidth : borderWidth;
}


void QuarticurveClient::activeChange()
{
	for(int i=QuarticurveClient::BtnHelp; i < QuarticurveClient::BtnCount; i++)
		if(button[i])
			button[i]->repaint(false);
		widget()->repaint(false);
}


void QuarticurveClient::resize( const QSize& s )
{
	widget()->resize( s );
}


QSize QuarticurveClient::minimumSize() const
{
	return widget()->minimumSize();
}


// The hiding button while shrinking, show button while expanding magic
void QuarticurveClient::calcHiddenButtons()
{
	// Hide buttons in this order:
	// Sticky, Help, Maximize, Minimize, Close, Menu.
	QuarticurveButton* btnArray[] = { button[BtnOnAllDesktops], button[BtnHelp],
		button[BtnMax], button[BtnIconify], button[BtnClose],
		button[BtnMenu] };

	int minwidth  = 160; // Start hiding at this width
	int btn_width = 16;
	int current_width = width();
	int count = 0;
	int i;

	// Find out how many buttons we need to hide.
	while (current_width < minwidth)
	{
		current_width += btn_width;
		count++;
	}

	// Bound the number of buttons to hide
	if (count > 6)
		count = 6;

	// Hide the required buttons...
	for(i = 0; i < count; i++)
	{
		if ( btnArray[i] && btnArray[i]->isVisible() )
			btnArray[i]->hide();
    }

	// Show the rest of the buttons...
	for(i = count; i < 6; i++)
	{
		if (btnArray[i] && (!btnArray[i]->isVisible()) )
			btnArray[i]->show();
	}
}


KDecoration::Position QuarticurveClient::mousePosition( const QPoint& p ) const
{
	Position m = PositionCenter;

	// Modify the mouse position if we are using a grab bar.
	if (showGrabBar && (!isTool()) )
		if (p.y() < (height() - 8))
			m = KDecoration::mousePosition(p);
		else
		{
			if (p.x() >= (width() - 20))
				m = PositionBottomRight;
			else if (p.x() <= 20)
				m = PositionBottomLeft;
			else
				m = PositionBottom;
		}
	else
		m = KDecoration::mousePosition(p);

	return m;
}


// Make sure the menu button follows double click conventions set in kcontrol
void QuarticurveClient::menuButtonPressed()
{
	static QTime t;
	static QuarticurveClient* lastClient = NULL;
	bool dbl = ( lastClient == this && t.elapsed() <= QApplication::doubleClickInterval());
	lastClient = this;
	t.start();

	if (dbl)
	{
		m_closing = true;
		return;
	}

	QPoint menupoint ( button[BtnMenu]->rect().bottomLeft().x()-1,
					button[BtnMenu]->rect().bottomLeft().y()+2 );
	KDecorationFactory* f = factory();
	showWindowMenu( button[BtnMenu]->mapToGlobal( menupoint ));
	if( !f->exists( this )) // 'this' was destroyed
		return;

	button[BtnMenu]->setDown(false);
}


void QuarticurveClient::menuButtonReleased()
{
	if (m_closing)
		closeWindow();
}


bool QuarticurveClient::eventFilter( QObject* o, QEvent* e )
{
	if ( o != widget() )
		return false;

	switch( e->type())
	{
		case QEvent::Resize:
			resizeEvent(static_cast< QResizeEvent* >( e ) );
			return true;
		case QEvent::Paint:
			paintEvent(static_cast< QPaintEvent* >( e ) );
			return true;
		case QEvent::MouseButtonDblClick:
			mouseDoubleClickEvent(static_cast< QMouseEvent* >( e ) );
			return true;
		case QEvent::MouseButtonPress:
			processMousePressEvent(static_cast< QMouseEvent* >( e ) );
			return true;
		case QEvent::Show:
			showEvent(static_cast<QShowEvent *>(e));
			return true;
		default:
			break;
	}
	return false;
}

} // namespace

// Extended KWindowSystem plugin interface
extern "C"  KDecorationFactory *create_factory()
{
	return new Quarticurve::QuarticurveHandler();
}

#include ".moc/moc_quarticurveclient.cpp"

// vim: ts=4
